/*
 * get.c
 *
 * By Ross Ridge
 * Public Domain
 *
 * Extract a requested delta from a SCCS file.
 * 
 */

#include "mysc.h"
#include "version.h"
#include "fileiter.h"
#include "sccsfile.h"
#include "pfile.h"
#include "getopt.h"

char const main_sccs_id[] = "@(#) MySC get.c 1.17 98/03/17 17:41:25";

/* Prints a list of included or excluded SIDs. */

static void
print_id_list(FILE *msgout, char const *s, list<sid> const &list) {
	int i, len;
	
	len = list.length();
	if (len > 0) {
#if 1
		/* Printing out the IDs all on one line instead of
                   one per line should make the output more clear */
		fprintf(msgout, "%s:", s);
		for(i = 0; i < len; i++) {
			putc(' ', msgout);
			list[i].print(msgout);
		}
		putc('\n', msgout);
#else
		/* Compatible output */
		fprintf(msgout, "%s\n", s);
		for(i = 0; i < len; i++) {
			list[i].print(msgout);
			putc('\n', msgout);
		}
#endif
	}
}

void
usage() {
	fprintf(stderr, msgptr(181,
"usage: %s [-begkmnpstV] [-c date] [-r SID] [-i range] [-w string]\n"
"\t[-x range] file ...\n"),
                prg_name);
}

int
do_get(FILE *msgout, sccs_pfile *pfile, sccs_name &name, string gname,
       sid rid, int latest, int branch, int use_stdout, int no_gfile,
       sccs_date cutoff, sid_list &include, sid_list &exclude, int keywords,
       char const *wstring, int show_sid, int show_module) {
	sccs_file file(name, sccs_file::READ);
	if (file.failed()) {
		return -1;
	}

	sid retrieve;
	if (latest) {
		 retrieve = file.find_latest_sid(rid);
	} else {
		 retrieve = file.find_requested_sid(rid);
	}

	if (retrieve == NULL) {
		error(msgptr(182, "%s: Requested SID not found."),
		      (char const *) name);
		return -1;
	}

	sid new_delta;
	if (pfile != NULL) {
		if (file.test_locks(retrieve, *pfile) == -1) {
			return -1;
		}
		new_delta = file.find_next_sid(rid, retrieve, branch, *pfile);
	}

	FILE *out = NULL;

	if (no_gfile) {
		out = open_null();
		gname = "null";
	} else if (use_stdout) {
		out = stdout;
		gname = "stdout";
	} else {
		assert(name.valid());

		if (gname == NULL) {
			gname = name.gfile();
		}

		int mode = CREATE_AS_REAL_USER | CREATE_FOR_GET;
		if (!keywords) {
			mode |= CREATE_READ_ONLY;
		}

		out = fcreate(gname, mode);

		if (out == NULL) {
			if (errno == EEXIST) {
				error(msgptr(183, "%s: File exists and"
					     " is writable."),
				      (char const *) name);
			} else {
				error(errno, msgptr(184, "%s: Can't open file"
						    " for writing"),
				      (char const *)gname);
			}
			return -1;
		}
	}
		
	struct sccs_file::get_status status
		= file.get(out, gname, retrieve, cutoff, include, exclude,
			   keywords, wstring, show_sid, show_module);

	if (!use_stdout) {
		if (fclose(out) == EOF) {
			error(errno, msgptr(5, "%s: error closing gfile"),
			      (char const *) gname);
			return -1;
		}
	}

	if (status.failed) {
		return -1;
	}

	fprintf(msgout, "\n%s\n", (char const *) name);

	print_id_list(msgout, msgptr(6, "Included"), status.included);
	print_id_list(msgout, msgptr(7, "Excluded"), status.excluded);
	retrieve.print(msgout);
	putc('\n', msgout);

	if (pfile != NULL) {
		fprintf(msgout, msgptr(8, "new delta "));
		new_delta.print(msgout);
		putc('\n', msgout);

		if (pfile->add_lock(retrieve, new_delta, include, exclude)
		    == -1) {
			return -1;
		}
	}

	if (!no_gfile) {
		fprintf(msgout, msgptr(9, "%d lines\n"), status.lines);
	}

	return 0;
}

int
main(int argc, char **argv) {
	int c;
	sid rid = NULL;			/* -r */
	int for_edit = 0;		/* -e */
	int branch = 0;			/* -b */
	int keywords = 0;		/* -k */
	int use_stdout = 0;		/* -p */
	int silent = 0;			/* -s */
	int no_gfile = 0;		/* -g */
	char const *wstring = NULL;	/* -w */
	sid_list include, exclude;	/* -i, -x */
	sccs_date cutoff;		/* -c */
	int show_sid = 0;		/* -m */
	int show_module = 0;		/* -n */
	string gname = NULL;		/* -G */
	int latest = 0;			/* -t */
#if 0
	int seq_no = 0;			/* -a */
#endif
	int return_status = 0;
	

	if (argc > 0) {
		set_prg_name(argv[0]);
	} else {
		set_prg_name("get");
	}

	setlocale(LC_ALL, "");
	init_msgdb();

	class getopt opts(argc, argv, "r:c:i:x:ebkl*pLsmngtw:a:VG:");
	for(c = opts.next(); c != getopt::END_OF_ARGUMENTS; c = opts.next()) {
		switch (c) {
		default:
			quit(-2, msgptr(10, "Unsupported option: '%c'"), c);

		case 'r':
			rid = sid(opts.getarg());
			if (!rid.valid()) {
				quit(-2, msgptr(11, "Invalid SID: '%s'"),
				     opts.getarg());
			}
			break;

		case 'c':
			cutoff = sccs_date(opts.getarg());
			if (!cutoff.valid()) {
				quit(-2,
				     msgptr(12, "Invalid cutoff date: '%s'"),
				     opts.getarg());
			}
			break;

		case 'i':
			include = sid_list(opts.getarg());
			if (!include.valid()) {
				quit(-2, 
				     msgptr(13, "Invalid inclusion list: '%s'"),
				     opts.getarg());
			}
			break;

		case 'x':
			exclude = sid_list(opts.getarg());
			if (!exclude.valid()) {
				quit(-2, 
				     msgptr(14, "Invalid exclusion list: '%s'"),
				     opts.getarg());
			}
			break;

		case 'e':
			for_edit = 1;
			keywords = 1;
			break;

		case 'b':
			branch = 1;
			break;

		case 'k':
			keywords = 1;
			break;

		case 'p':
			use_stdout = 1;
			break;

		case 's':
			silent = 1;
			break;

		case 'm':
			show_module = 1;
			break;

		case 'n':
			show_sid = 1;
			break;

		case 'g':
			no_gfile = 1;
			break;

		case 'w':
			wstring = opts.getarg();
			break;

#if 0		       
		case 'a':
			int i = atoi(opts.getarg());
			if (i < 1) {
				quit(-2,
				     msgptr(15,
					    "Invalid sequence number: '%s'"),
				     optarg);
			}
			seq_no = i;
			break;
#endif
		case 'G':
			gname = opts.getarg();
			break;

		case 't':
			latest = 1;
			break;

		case 'V':
			version();
			break;
		}
	}

	FILE *msgout = stdout;

	if (use_stdout) {
		msgout = stderr;
	}

	if (silent) {
		msgout = open_null();
		if (msgout == NULL) {
			return 1;
		}
	}

	sccs_file_iterator iter(argc, argv, opts.get_index());
	if (iter.length_zero()) {
		quit(-2, msgptr(16, "No SCCS file specified."));
	}

	if (gname != NULL && iter.length_many()) {
		quit(-2, msgptr(185, "The -G option may only be used with"
				   " one SCCS file."));
	}

	for(; iter != sccs_file_iterator(); ++iter) {
		sccs_name &name = *iter;
		sccs_pfile *pfile = NULL;

		if (for_edit) {
			pfile = new sccs_pfile(name, sccs_pfile::APPEND);
			if (pfile->failed()) {
				delete pfile;
				return_status = 1;
				continue;
			}
		}
		
		if (do_get(msgout, pfile, name, gname, rid, latest, branch,
			   use_stdout, no_gfile, cutoff, include, exclude,
			   keywords, wstring, show_sid, show_module)
		    == -1) {
			return_status = 1;
		}

		if (pfile != NULL) {
			delete pfile;
		}
	}

	return return_status;
}

/* Local variables: */
/* mode: c++ */
/* End: */
