/*
 * pfile.c
 *
 * By Ross Ridge
 * Public Domain
 *
 * Common members of the class sccs_pfile.
 *
 */

#include "mysc.h"
#include "linebuf.h"
#include "pfile.h"

#ifdef CONFIG_SCCS_IDS
static char const sccs_id[] = "@(#) MySC pfile.c 1.5 98/01/06 08:24:29";
#endif

void
sccs_pfile::corrupt(int lineno, char const *msg) const
{
	error(msgptr(51, "%s: line %d: p-file is corrupt.  (%s)"),
	      (char const *) pname, lineno, msg);
}

sccs_pfile::sccs_pfile(sccs_name &n, enum _mode m)
		: name(n), mode(m), pos(-1), error_status(0) {

        if (!name.valid()) {
		error(msgptr(20, "%s: Not an SCCS file."), (char const *) name);
		mode = READ;
		error_status = 1;
		return;
	}

	pname = name.pfile();

	if (mode != READ) {
		if (name.lock()) {
			error(msgptr(52, "%s: SCCS file is locked."
				     "  Try again later."),
			      (char const *) name);
			error_status = 1;
			return;
		}
	}

	FILE *pf = fopen(pname, "rb");
	if (pf == NULL) {
		if (errno != ENOENT) {
			error(msgptr(53, "%s: Can't open file for reading."),
			      (char const *) pname);
			error_status = 1;
			return;
		}
	} else {
		class _linebuf linebuf;

		int lineno = 0;
		while(!linebuf.read_line(pf)) {
			linebuf[strlen(linebuf) - 1] = '\0';
			lineno++;

			char *args[7];

			int argc = split(linebuf, args, 7, ' ');

			if (argc < 5) {
				corrupt(lineno,
					msgptr(54, "Expected 5-7 args"));
				error_status = 1;
				return;
			}

			char *incl = NULL, *excl = NULL;
			int i;
			for(i = 5; i < argc; i++) {
				if (args[i][0] == '-') {
					if (args[i][1] == 'i') {
						incl = args[i] + 2;
					} else if (args[i][1] == 'x') {
						excl = args[i] + 2;
					}
				}
			}

			if ((argc == 6 && excl == NULL && incl == NULL)
			    || (argc == 7 && (excl == NULL || incl == NULL))) {
				corrupt(lineno,
					msgptr(55, "Bad -i or -x option"));
				error_status = 1;
				return;
			}
				      
				
			struct edit_lock tmp(args[0], args[1], args[2],
					     args[3], args[4], 
					     incl, excl);

			if (!tmp.got.valid() || !tmp.got.valid()) {
				corrupt(lineno, msgptr(56, "Invalid SID"));
				error_status = 1;
				return;
			}

			if (!tmp.date.valid()) {
				corrupt(lineno, msgptr(57, "Invalid date"));
				error_status = 1;
				return;
			}
					
			if (!tmp.include.valid() || !tmp.exclude.valid()) {
				corrupt(lineno, msgptr(58, "Invalid SID list"));
				error_status = 1;
				return;
			}
			edit_locks.add(tmp);
		}

		if (ferror(pf)) {
			error(errno, msgptr(59, "%s: Read error."),
			      (char const *) pname);
			error_status = 1;
			return;
		}

		fclose(pf);
	}
}

int
sccs_pfile::is_locked(sid id) {
	rewind();

	sccs_pfile &it = *this;
	while(next()) {
		if (it->got == id) {
			return 1;
		}
	}

	return 0;
}

int
sccs_pfile::is_to_be_created(sid id) {
	rewind();

	sccs_pfile &it = *this;
	while(next()) {
		if (it->delta == id) {
			return 1;
		}
	}

	return 0;
}

sccs_pfile::~sccs_pfile() {
	if (mode != READ) {
		name.unlock();
	}
}

/* Local variables: */
/* mode: c++ */
/* End: */
