/*
 * pipe.c
 *
 * By Ross Ridge
 * Public Domain
 *
 * Members of the class pipe.
 *
 */

#include "mysc.h"
#include "pipe.h"
#include "run.h"
#include "list.h"
#include "sysdep.h"

#ifdef CONFIG_SCCS_IDS
static char const sccs_id[] = "@(#) MySC pipe.c 1.11 98/06/06 16:25:53";
#endif

extern int create(string name, int mode); /* file.c */

string
make_tempname() {
	string name;

#ifdef CONFIG_TEMP_DIRECTORY
	char *s = tempnam(CONFIG_TEMP_DIRECTORY, "mysc");
	name = s;
	if (s == NULL) {
		error(msgptr(60, "tempnam() failed."));
		return NULL;
	}
	free(s);
#else
	name = tmpnam(NULL);
	if (name == NULL) {
		error(msgptr(61, "tmpnam() failed."));
		return NULL;
	}
#endif
	return name;
}

#ifdef CONFIG_NO_PIPE

/* Deletes the temporary file if the programme quits prematurely. */

int
pipe::do_cleanup() {
	if (f != NULL) {
		fclose(f);
		f = NULL;
		fd = -1;
	}
	if (fd != -1) {
		close(fd);
		fd = -1;
	}
	if (name != NULL) {
		remove(name);
		name = NULL;
	}
	return 0;
}

/* Constructor for class pipe.  Creates a temporary file to emulate
   a pipe. */

pipe::pipe(): f(NULL), fd(-1), name(NULL), failed(0) {
	name = make_tempname();
	if (name == NULL) {
		failed = 1;
	}

	fd = create(name, CREATE_EXCLUSIVE | CREATE_READ_ONLY);
	if (fd == -1) {
		error(errno, msgptr(47, "%s: Can't create temporary file."),
		      (char const *) name);
		failed = 1;
		return;
	}
	f = fdopen(fd, "w");
	if (f == NULL) {
		error(errno, msgptr(62, "fdopen() failed."));
		failed = 1;
	}
}


/* Returns the stream to use to read from the pipe. */

FILE *
pipe::read_stream() {
	fd = open(name, O_RDONLY);
	if (fd == -1) {
		error(errno, msgptr(63, "%s: Can't reopen temporary file."),
		      (char const *) name);
		failed = 1;
		return NULL;
	}
	f = fdopen(fd, "r");
	if (f == NULL) {
		error(errno, msgptr(62, "fdopen() failed."));
		close(fd);
		fd = -1;
		failed = 1;
		return NULL;
	}
	return f;
}


/* Closes the read stream and deletes the temporary file. */

int
pipe::read_close() {
	if (failed) {
		return -1;
	}

	fclose(f);
	f = NULL;
	fd = -1;

	if (remove(name) == -1) {
		error(errno, msgptr(49, "%s: Can't remove temporary file."),
		      (char const *) name);
		failed = 1;
		return -1;
	}
	name = NULL;
	return 0;
}

/* Runs the diff command with a one pipe as stdin and another pipe
   as stdout. */

int	
run_diff(char const *gname, pipe *in, pipe *out, list<string> const &args) {
	if (in->failed || (out != NULL && out->failed)) {
		return -1;
	}

	int old_stdin = dup(0);
	if (old_stdin == -1) {
		error(errno, msgptr(64, "dup(0) failed."));
		in->failed = 1;
		if (out != NULL) {
			out->failed = 1;
		}
		return -1;
	}

	int old_stdout;
	if (out != NULL) {
		old_stdout = dup(1);
		if (old_stdout == -1) {
			error(errno, msgptr(33, "dup(1) failed."));
			close(old_stdin);
			in->failed = out->failed = 1;
			return -1;
		}
	}

	in->read_stream();

	close(0);
	if (dup(in->fd) != 0) {
		quit(errno, msgptr(65, "dup() != 0."));
	}

	if (out != NULL) {
		close(1);
		if (dup(out->fd) != 1) {
			quit(errno, msgptr(66, "dup() != 1."));
		}
	}

	list<string> new_args = args;
	new_args.add("-");
	new_args.add(gname);

	int ret = run(CONFIG_DIFF_COMMAND, new_args);

	close(0);
	if (dup(old_stdin) != 0) {
		quit(errno, msgptr(67, "can't restore stdin/stdout"));
	}
	close(old_stdin);

	if (old == NULL) {
		close(1);
		if (dup(old_stdout) != 1) {
			quit(errno, msgptr(67, "can't restore stdin/stdout"));
		}
		close(old_stdout);
	}

	return ret;
}


#else /* CONFIG_NO_PIPE */

#ifdef CONFIG_NO_FORK
#error "CONFIG_NO_FORK may not be defined if CONFIG_NO_PIPE isn't defined."
#endif


/* Head of list class wait_pid objects that haven't been reaped yet. */

class wait_pid *wait_pid::head = NULL;


/* Removes "this" from the list of unreaped wait_pid's */

void
wait_pid::unlink() {
	if (!linked) {
		return;
	}

	class wait_pid *p = head;

	if (p == this) {
		head = next;
		linked = 0;
		return;
	}

	while(p != NULL) {
		if (p->next == this) {
			p->next = next;
			linked = 0;
			return;
		}
		p = p->next;
	}

	assert(&wait_pid::unlink);
}


/* Waits for until the this pid's process ends and then returns its
   exit status.  Records the exit status of any other processes
   that exits in the mean time. */

int
wait_pid::wait() {
	assert(pid != -1 && pid != 0);

	if (reaped) {
		pid = -1;
		return status;
	}

	int st;
	int r = ::wait(&st);
	while(r != pid) {
		if (r != -1) {
			class wait_pid *p = head;

			while(p != NULL) {
				if (p->pid == r) {
					unlink();
					p->reaped = 1;
					p->status = st;
					break;
				}
				p = p->next;
			}
		} else if (errno != EINTR) {
			quit(errno, msgptr(68, "wait() failed."));
		}
		r = ::wait(&st);
	}

	unlink();
	pid = -1;

	return st;
}


/* This is used to keep system dependencies out of the include files. */
void
pipe::_exit(int ret) {
	::_exit(ret);
}


/* Constructor for the class pipe.  Creates a pipe and forks, giving
   the write end of the pipe to child process and the read end to the
   parent process. */

FILE *
pipe::write_stream() {
	if (failed) {
		return NULL;
	}

	int fds[2];

	if (::pipe(fds) == -1) {
		error(errno, msgptr(69, "pipe() failed."));
		failed = 1;
		return NULL;
	}

	pid = fork();
	if (pid == -1) {
		error(errno, msgptr(70, "fork() failed."));
		failed = 1;
		return NULL;
	}

	if (pid == 0) {
		child = 1;
		cleanup::set_in_child();
		close(fds[0]);
		fd = fds[1];
		f = fdopen(fds[1], "w");
		if (f == NULL) {
			quit(errno, 4, msgptr(62, "fdopen() failed."));
		}
		return f;
	} else {
		child = 0;
		close(fds[1]);
		fd = fds[0];
		f = fdopen(fds[0], "r");
		if (f == NULL) {
			error(errno, msgptr(62, "fdopen() failed."));
			failed = 1;
		}
		return NULL;
	}
}


/* Runs the diff command with a one pipe as stdin and another pipe
   as stdout. */

int
run_diff(char const *gname, class pipe *in, class pipe *out,
	 list<string> const &args) {

	if (in->failed) {
		return -1;
	}
	if (out == NULL) {
		wait_pid pid;
		pid = fork();
		if (pid == -1) {
			perror("fork");
			return -1;
		}
		if (pid != 0) {
			return pid.wait();
		}
		cleanup::set_in_child();
	} else {
		if (out->failed) {
			return -1;
		}
		if (out->write_stream() == NULL) {
			return 0;
		}
	}

	int len = args.length();
	int i;
	char const **argv = (char const **) xmalloc((len + 4)
						    * sizeof(char const **));

	argv[0] = CONFIG_DIFF_COMMAND;

	for(i = 0; i < len; i++) {
		argv[i + 1] = args[i];
	}
	     
	argv[i + 1] = "-";
	argv[i + 2] = gname;
	argv[i + 3] = NULL;
	
	close(0);
	if (dup(in->fd) != 0) {
		quit(errno, 4, msgptr(65, "dup() != 0."));
	}

	if (out != NULL) {
		close(1);
		if (dup(out->fd) != 1) {
			quit(errno, 4, msgptr(66, "dup() != 1."));
		}
	}

	execv(CONFIG_DIFF_COMMAND, (char **) argv);

	quit(errno, 127, msgptr(71, "execl(\"%s\") failed."),
	     CONFIG_DIFF_COMMAND);
}

#endif /* CONFIG_NO_PIPE */

/* Local variables: */
/* mode: c++ */
/* End: */


