/*
 * pipe.h
 *
 * By Ross Ridge
 * Public Domain
 *
 * Defines the class pipe.
 *
 * @(#) MySC pipe.h 1.10 98/06/06 16:10:23
 *
 */

#ifndef __PIPE_H__
#define __PIPE_H__

#include "list.h"

/* One of two definitions of the class pipe are used depending if the
   system supports pipes (and fork) or not.  If pipes are not supported
   then a temporary file used to emulate one.  The members must be
   called in order: constructor, write_stream(), write_close(),
   read_stream(), read_close(), destructor. */

#ifdef CONFIG_NO_PIPE

class pipe: cleanup {
	friend int run_diff(char const *gname, pipe *in, pipe *out,
			    list<string> const &args);

	FILE *f;
	int fd;
	string name;

	int do_cleanup();
	int failed;

public:
	pipe();

	FILE *write_stream() { return f; }

	int
	write_close() { 
		if (!failed && fclose(f) == EOF) {
			error(msgptr(42, "%s: Write error."), name);
			failed = 1;
			return -1;
		}
		return 0;
	}

	FILE *read_stream();
	int read_close();

	void set_failed() { failed = 1; }
	int get_failed() { return failed; }

	~pipe() { 
		assert(failed || (f == NULL && name == NULL));
		if (f != NULL) {
			fclose(f);
		}
		if (name != NULL) {
			remove(name);
		}
	}
};

#else /*  CONFIG_NO_PIPE */

/* Definition of the class wait_pid.  It handles the problem of
   other child processes exiting while waiting a for a specific
   child process to exit. */

class wait_pid {
	int pid;
	int reaped;
	int status;
	int linked;
	class wait_pid *next;
	static class wait_pid *head;

	void
	link(int p) {
		if (p != -1 && p != 0 && !linked) {
			linked = 1;
			reaped = 0;
			next = head;
			head = this;
		}
		pid = p;
	}

	void unlink();

	void operator =(wait_pid const &);
	wait_pid(wait_pid const &);

public:
	wait_pid(): pid(-1), linked(0) {};

	wait_pid(int p): linked(0) { link(p); }

	int
	operator =(int p) {
		assert(pid == -1);
		link(p);
		return p;
	}

	operator int() {
		return pid;
	}

	int wait();

	~wait_pid() {
		if (linked) {
			unlink();
		}
	}
};

class pipe {
	friend int run_diff(char const *gname,
			    class pipe *in, class pipe *out,
			    list<string> const &args);

	int fd;
	wait_pid pid;
	int child;
	FILE *f;
	int failed;

	static void _exit(int) NORETURN;

	pipe(class pipe const &) { abort(); };
	void operator =(class pipe const &) { abort(); };
		   
public:
	pipe(): child(0), f(NULL), failed(0) {};

	FILE *write_stream();

	int
	write_close() {
		if (failed) {
			if (child) {
				_exit(1);
			}
			return -1;
		}
		if (child) {
			if (fclose(f) == EOF) {
				error(errno, msgptr(42, "%s: Write error."),
				      "pipe");
				_exit(1);
			} else {
				_exit(0);
			}
		}
		return 0;
	}

	FILE *
	read_stream() {
		assert(!child);
		if (failed) {
			return NULL;
		}
		return f;
	}

	int
	read_close() {
		assert(!child);
		if (failed) {
			return -1;
		}
		fclose(f);
		f = NULL;
		return pid.wait();
	}

	void set_failed() { failed = 1; }
	int get_failed() { return failed; }

	~pipe() {
		assert(!child);
		assert(failed || f == NULL);
		if (f != NULL) {
			fclose(f);
		}
	}
};	

#endif /* CONFIG_NO_PIPE */

string make_tempname();

#endif /* __PIPE_H__ */

/* Local variables: */
/* mode: c++ */
/* End: */
