/*
 * sccsdiff.c
 *
 * By Ross Ridge
 * Public Domain
 *
 * Compare SCCS file revisions.
 * 
 */

#include "mysc.h"
#include "fileiter.h"
#include "sccsfile.h"
#include "version.h"
#include "pipe.h"
#include "run.h"
#include "getopt.h"

/*
 * All the diff options from GNU diff and Solaris diff that make
 * sense when used with sccsdiff.
 */
#define DIFF_OPTS "abcdenfhilnpqstuwyw:vBC:D:I:HNF:U:"

char const main_sccs_id[] = "@(#) MySC sccsdiff.c 1.5 00/05/12 16:11:27";

static int
do_diff(sccs_file &file, sid rid0, sid rid1, FILE *gfile, string gname,
	list<string> const &args, int keywords) {
	struct sccs_file::get_status status;

	if (rid1 != NULL) {
	 	status = file.get(gfile, gname, rid1, NULL, NULL, NULL,
				  0, NULL, 0, 0);
		if (status.failed) {
			return -1;
		}
	}

	class pipe diff_in;
	FILE *get_out = diff_in.write_stream();
	if (get_out != NULL) {
		status = file.get(get_out, msgptr(131, "(diff pipe)"), rid0,
			 	  NULL, NULL, NULL, keywords, NULL, 0, 0);
		if (status.failed) {
			diff_in.set_failed();
		}
	}
	diff_in.write_close();

	int diff_ret = run_diff(gname, &diff_in, NULL, args);
	if (diff_ret == -1) {
		diff_in.set_failed();
	} else if (diff_ret != STATUS(0) && diff_ret != STATUS(1)) {
		error(msgptr(189, "diff: Subprocess exited abnormally."
			          "  (status = %d)"), STATUS_MSG(diff_ret));
		diff_in.set_failed();
	}

	int ret = diff_in.read_close();
	if (ret != -1 && ret != STATUS(0)) {
		error(msgptr(133, "get: Subprocess exited abnormally."
			          "  (status = %d)"), STATUS_MSG(ret));

		diff_in.set_failed();
	}

	if (diff_in.get_failed()) {
		return -1;
	}

	return diff_ret;
}

static void
add_rid(list<sid> &rids, char const *arg) {
	if (rids.length() == 2) {
		quit(-2, msgptr(187, "Only two SIDs may be specified"));
	}
	sid rid(arg);
	if (!rid.valid()) {
		quit(-2, msgptr(11, "Invalid SID: '%s'"), arg);
	}
	rids.add(rid);
}

void
usage() {
	fprintf(stderr, msgptr(188,
"usage: %s [-R] [-r SID [-r SID]] [-k] [diff-opts] file ...\n"),
                prg_name);
}

int
main(int argc, char **argv) {
	list<sid> rids;
	list<string> diff_opts;
	char const *s;
	char opt[3] = "-?";
	int return_status = 0;
	int keywords = 0;

	if (argc > 0) {
		set_prg_name(argv[0]);
	} else {
		set_prg_name("sccsdiff");
	}

	setlocale(LC_ALL, "");
	init_msgdb();

	int c;
	class getopt opts(argc, argv, "r:RVk" DIFF_OPTS);
	for(c = opts.next(); c != getopt::END_OF_ARGUMENTS; c = opts.next()) {
		switch (c) {
		default:
			s = strchr(DIFF_OPTS, c);
			if (s == NULL) {
				quit(-2, msgptr(10, "Unsupported option: '%c'"),
				     c);
			}
			opt[1] = c;
			diff_opts.add(string(opt));
			if (s[1] == ':') {
				diff_opts.add(opts.getarg());
			}
			break;

		case 'r':
			add_rid(rids, opts.getarg());
			break;

		case 'R':
			add_rid(rids, NULL);
			break;

		case 'k':
			keywords = 1;
			break;

		case 'V':
			version();
			break;
		}
	}

	sccs_file_iterator iter(argc, argv, opts.get_index());
	if (iter.length_zero()) {
		quit(-2, msgptr(16, "No SCCS file specified."));
	}

	for(; iter != sccs_file_iterator(); ++iter) {
		sccs_name &name = *iter;
		sid rid0, rid1;
		string gname;
		FILE *gfile;

		sccs_file file(name, sccs_file::READ);
		if (file.failed()) {	
			return_status = 2;
			continue;
		}


		switch(rids.length()) {
		case 0:
			rid0 = file.find_requested_sid(NULL);
			rid1 = NULL;
			break;

		case 1:
			rid0 = file.find_requested_sid(rids[0]);
			rid1 = NULL;
			break;

		case 2:
			rid0 = file.find_requested_sid(rids[0]);
			rid1 = file.find_requested_sid(rids[1]);
			break;
		}

		if (rid0 == NULL || (rids.length() == 2 && rid1 == NULL)) {
			error(msgptr(182, "%s: Requested SID not found."),
			      (char const *) name);
			return_status = 2;
			continue;
		}

		if (rid1 == NULL) {
			gname = name.gfile();
			gfile = NULL;
		} else {
			gname = make_tempname();
			gfile = fcreate(gname, CREATE_EXCLUSIVE);
			if (gfile == NULL) {
				error(errno,
				      msgptr(175, "%s: Can't open file."),
				      (char const *) gname);
				return_status = 2;
				continue;
			}
		}

		int r = do_diff(file, rid0, rid1, gfile, gname, diff_opts,
			        keywords);
		if (r == -1) {
			return_status = 2;
		} else if (r == 1) {
			if (return_status != 2) {
				return_status = 1;
			}
		} else if (r != 0) {
			return_status = r;
		}

		if (gfile != NULL) {
			fclose(gfile);
		}

		if (rid1 != NULL && remove(gname) == -1) {
			error(errno,
			      msgptr(49, "%s: Can't remove temporary file."),
			      (char const *) gname);
			return_status = 2;
		}
	}	

	return return_status;
}

/* Local variables: */
/* mode: c++ */
/* End: */
