/*
 * sccsfile.c 
 *
 * By Ross Ridge
 * Public Domain
 *
 * Common members of the class sccs_file and its subclasses.  Most of
 * the members in this file are used to read from the SCCS file.
 *
 */

#include "mysc.h"
#include "sccsfile.h"

#include <ctype.h>

#ifdef CONFIG_SCCS_IDS
static char const sccs_id[] = "@(#) MySC sccsfile.c 1.12 00/05/12 14:43:33";
#endif

/* struct */ sccs_file::delta &
sccs_file::delta::operator =(struct delta const &it) {
	inserted = it.inserted;
	deleted = it.deleted;
	unchanged = it.unchanged;
	type = it.type;
	id = it.id;
	date = it.date;
	user = it.user;
	seq = it.seq;
	prev_seq = it.prev_seq;
	included = it.included;
	excluded = it.excluded;
	ignored = it.ignored;
	mrs = it.mrs;
	comments = it.comments;
	return *this;
}


int
sccs_file::_delta_table::update_highest(struct delta const &it) {
	seq_no seq = it.seq;
	if (seq > high_seqno) {
		if (seq_table != NULL) {
			seq_table = (int *) xrealloc(seq_table, 
						     (seq + 1) * sizeof(int));
			int i;
			for(i = high_seqno + 1; i < seq + 1; i++) {
				seq_table[i] = -1;
			}
		}
		high_seqno = it.seq;
	}

	if (it.type == 'D' && it.id > high_release) {
		high_release = it.id;
	}

	if (seq_table != NULL) {
		if (seq_table[seq] != -1) {
			error(msgptr(85, "Sequence number %u is duplicated"
				     " in delta table."), seq);
			return -1;
		}
		seq_table[seq] = length() - 1;
	}

	return 0;
}


/* Adds a delta to the end of the delta_table. */

int
sccs_file::_delta_table::add(struct delta const &it) {
	list<struct delta>::add(it);
	return update_highest(it);
}


/* Finds a delta in the delta table by its SID. */

/* struct */ sccs_file::delta const *
sccs_file::_delta_table::find(sid id) const {
	delta_iterator iter(*this);

	while(iter.next()) {
		if (iter->id == id) {
			return iter.operator->();
		}
	}

	return NULL;
}

/* Static member for opening a SCCS file and then calculating its checksum. */

FILE *
sccs_file::open_sccs_file(char const *name, enum _mode mode, unsigned *sump) {
	FILE *f;

	if (mode == UPDATE) {
		f = fopen(name, "rb+");
	} else {
		f = fopen(name, "rb");
	}

	if (f == NULL) {
		error(errno,
		      msgptr(86, "%s: Can't open SCCS file for reading."),
		      name);
		return NULL;
	}

	if (getc(f) != '\001' || getc(f) != 'h') {
		error(msgptr(87, "%s: Bad magic number"), name);
		fclose(f);
		return NULL;
	}

	int c = getc(f);
	while(c != '\n') {
		if (c == EOF) {
			error(errno, msgptr(88, "%s: Unexpected EOF."), name);
			fclose(f);
			return NULL;
		}
		c = getc(f);
	}

	unsigned sum = 0;
	c = getc(f);
	while(c != EOF) {
		sum += (unsigned) c & 0xFF;
		c = getc(f);
	}

	if (ferror(f)) {
		error(errno, msgptr(59, "%s: Read error."),
		      (char const *) name);
		fclose(f);
		return NULL;
	}

	*sump = sum & 0xFFFF;

	rewind(f);

	return f;
}


/* Reads a line from the SCCS file and increments the current line number.
   If the end of file is reached it returns -1.  If it's a control line
   (it starts with ^A) then the control character (the second character)
   is returned.  Otherwise 0 is returned. */

int
sccs_file::read_line() {
	if (read_line_param(f)) {
		if (ferror(f)) {
			error(errno, msgptr(59, "%s: Read error."),
			      (char const *) name);
			return -2;
		}
		return -1;
	} 

	lineno++;
	if (linebuf[0] == '\001') {
#if 0
		fprintf(stderr, "@%s\n", (char const *) linebuf + 1);
#endif
		return linebuf[1];
	}
#if 0
	fprintf(stderr, ":%s\n", (char const *) linebuf);
#endif
	return 0;

}


/* Prints a message saying that SCCS file is corrupt. */

void
sccs_file::corrupt(char const *why) const {
	error(msgptr(89, "%s: line %d: Corrupted SCCS file. (%s)"),
	      (char const *) name, lineno, why);
}


/* Checks that a control line has at least one argument. */

int
sccs_file::check_arg() const {
	if (linebuf[2] != ' ') {
		corrupt(msgptr(90, "Missing arg"));
		return -1;
	}
	return 0;
}


/* Checks the a control line has no arguments. */

int
sccs_file::check_noarg() const {
	if (linebuf[2] != '\0') {
		corrupt(msgptr(91, "Unexpected arg"));
		return -1;
	}
	return 0;
}


/* Converts an ASCII string to an unsigned short, quiting if the
   string isn't a valid number. */

unsigned short
sccs_file::strict_atous(char const *s, int *err) const {
	long n = 0;

	char c = *s++;
	while(c != '\0') {
		if (!isdigit(c)) {
			corrupt(msgptr(92, "Invalid number"));
			*err = 1;
			return 65535;
		}
		n = n * 10 + (c - '0');
		if (n > 65535L) {
			corrupt(msgptr(93, "Number too big"));
			*err = 1;
			return 65535;
		}
		c = *s++;
	}

	return (unsigned short) n;
}

/* Reads a delta from the SCCS file's delta table and adds it to the
   delta table. */

int
sccs_file::read_delta() {
	/* The current line should be an 's' control line */

	assert(linebuf[1] == 's');
	if (check_arg() == -1) {
		return -1;
	}
	
	char *args[7];		/* Stores the result of spliting a line */

	if (split(linebuf + 3, args, 3, '/') != 3) {
		corrupt(msgptr(94, "Two /'s expected"));
		return -1;
	}

	struct delta tmp;	/* The new delta */

	int err = 0;
	tmp.inserted = strict_atous(args[0], &err);
	tmp.deleted = strict_atous(args[1], &err);
	tmp.unchanged = strict_atous(args[2], &err);
	if (err) {
		return -1;
	}

	if (read_line() != 'd') {
		corrupt(msgptr(95, "Expected '@d'"));
		return -1;
	}

	if (check_arg() == -1) {
		return -1;
	}

	split(linebuf + 3, args, 7, ' ');

	tmp.type = args[0][0];
	tmp.id = sid(args[1]);
	tmp.date = sccs_date(args[2], args[3]);
	tmp.user = args[4];
	tmp.seq = strict_atous(args[5], &err);
	tmp.prev_seq = strict_atous(args[6], &err);
	if (err) {
		return -1;
	}

	if ((tmp.type != 'R' && tmp.type != 'D') || args[0][1] != '\0') {
		corrupt(msgptr(96, "Bad delta type"));
		return -1;
	}
	if (!tmp.id.valid()) {
		corrupt(msgptr(56, "Invalid SID"));
		return -1;
	}
	if (!tmp.date.valid()) {
		corrupt(msgptr(97, "Bad Date/Time"));
		return -1;
	}

	/* Read in any lists of included, excluded or ignored seq. no's. */

	int c = read_line();
	int i;
 	char *start;
	for(i = 0; i < 3; i++) {
		if (c == "ixg"[i]) {
			if (check_arg() == -1) {
				return -1;
			}

			start = linebuf + 3;
			do {
				char *end = strchr(start, ' ');
				if (end != NULL) {
					*end++ = '\0';
				}
				seq_no seq = strict_atous(start, &err);
				if (err) {
					return -1;
				}

				switch(c) {
				case 'i':
					tmp.included.add(seq);
					break;

				case 'x':
					tmp.excluded.add(seq);
					break;

				case 'g':
					tmp.ignored.add(seq);
					break;
				}
				start = end;
			} while(start != NULL);

			c = read_line();
		}
	}

	while(c == 'm') {
		if (check_arg() == -1) {
			return -1;
		}
		tmp.mrs.add(linebuf + 3);
		c = read_line();
	}

	while(c == 'c') {
		if (check_arg() == -1) {
			return -1;
		}
		tmp.comments.add(linebuf + 3);
		c = read_line();
	}

	if (c != 'e') {
		corrupt(msgptr(98, "Expected '@e'"));
		return -1;
	}

	if (check_noarg() == -1) {
		return -1;
	}

	return delta_table.add(tmp);
}


/* Seeks on the SCCS file to the start of the body.  This function 
   may be rewritten as fseek() doesn't always work too well on
   text files. */

int
sccs_file::seek_to_body() {
	if (fseek(f, body_offset, 0) != 0) {
		error(errno, msgptr(99, "%s: fseek() failed."),
		      (char const *) name);
		return -1;
	}
	lineno = body_lineno;
	return 0;
}



/* Returns the module name of the SCCS file. */

string
sccs_file::get_module_name() const {
	if (flags.module == NULL) {
		return name.gfile();
	}
	return flags.module;
}

/* Constructor for the class sccs_file.  Unless the SCCS file is being
   created it reads in the all but the body of the file.  The file is
   locked if it isn't only being read.  */

sccs_file::sccs_file(sccs_name &n, enum _mode m)
	: name(n), f(NULL), mode(m), lineno(0) {

	fail_status = 1;

	if (!name.valid()) {
		error(msgptr(20, "%s: Not an SCCS file."), (char const *) name);
		mode = READ;
		return;
	}

	flags.branch = 0;
	flags.floor = NULL;
	flags.ceiling = NULL;
	flags.default_sid = NULL;
	flags.null_deltas = 0;
	flags.joint_edit = 0;
	flags.all_locked = 0;
		
	if (mode != READ) {
		if (name.lock()) {
			error(msgptr(52, "%s: SCCS file is locked."
				     "  Try again later."),
			      (char const *) name);
			return;
		}
	}

	if (mode == CREATE) {
		fail_status = 0;
		return;
	}

	unsigned sum;
	f = open_sccs_file(name, READ, &sum);
	if (f == NULL) {
		return;
	}

	int c = read_line();
	assert(c == 'h');

	int err = 0;
	unsigned short tmp = strict_atous(linebuf + 2, &err);
	if (err) {
		return;
	}
	if (tmp != sum) {
		fprintf(stderr, msgptr(100, "%s: Warning bad checksum.\n"),
			(char const *) name);
	}
	
	c = read_line();
	while(c == 's') {
		if (read_delta() == -1) {
			return;
		}
		c = read_line();
	}

	if (c != 'u') {
		corrupt(msgptr(101, "Expected '@u'"));
		return;
	}

	if (check_noarg() == -1) {
		return;
	}

	c = read_line();
	while(c != 'U') {
		if (c != 0) {
			corrupt(msgptr(102, "User name expected."));
			return;
		}
		users.add((char const *)linebuf);
		c = read_line();
	}

	if (check_noarg() == -1) {
		return;
	}

	c = read_line();
	while(c == 'f') {
		if (check_arg() == -1) {
			return;
		}

		if (linebuf[3] == '\0'
		    || (linebuf[4] != '\0' && linebuf[4] != ' ')) {
			corrupt(msgptr(103, "Bad flag arg."));
			return;
		}

		char *arg = NULL;
		if (linebuf[4] == ' ') {
			arg = linebuf + 5;
		}

		switch(linebuf[3]) {
		case 't':
			flags.type = arg;
			break;

		case 'v':
			flags.mr_checker = arg;
			break;

		case 'i':
			if (arg == NULL) {
				flags.id_keywords = "";
			} else {
				flags.id_keywords = arg;
			}
			break;

		case 'b':
			flags.branch = 1;
			break;

		case 'm':
			flags.module = arg;
			break;

		case 'f':
			flags.floor = release(arg);
			if (!flags.floor.valid()) {
				corrupt(msgptr(104, "Bad 'f' flag"));
				return;
			}
			break;

		case 'c':
			flags.ceiling = release(arg);
			if (!flags.ceiling.valid()) {
				corrupt(msgptr(105, "Bad 'c' flag"));
				return;
			}
			break;

		case 'd':
			flags.default_sid = sid(arg);
			if (!flags.default_sid.valid()) {
				corrupt(msgptr(106, "Bad 'd' flag"));
				return;
			}
			break;

		case 'n':
			flags.null_deltas = 1;
			break;

		case 'j':
			flags.joint_edit = 1;
			break;

		case 'l':
			if (arg != NULL && strcmp(arg, "a") == 0) {
				flags.all_locked = 1;
			} else {
				flags.locked = release_list(arg);
			}
			break;

		case 'q':
			flags.user_def = arg;
			break;

		case 'z':
			flags.reserved = arg;
			break;

		case 'e':
			/* Ignore.  Hopefully this isn't a uuencoded
			   SCCS archive. */
			break;

		default:
			corrupt(msgptr(107, "Unknown flag."));
			return;
		}

		c = read_line();
	}

	if (c != 't') {
		corrupt(msgptr(108, "Expected '@t'"));
		return;
	}

	if (check_noarg() == -1) {
		return;
	}

	c = read_line();
	while(c == 0) {
		comments.add((char const *)linebuf);
		c = read_line();
	}

	if (c != 'T') {
		corrupt(msgptr(109, "Expected '@T'"));
		return;
	}

	if (check_noarg() == -1) {
		return;
	}

	body_offset = ftell(f);
	if (body_offset == -1L) {
		error(errno, msgptr(110, "ftell() failed."));
		return;
	}

	body_lineno = lineno;

	fail_status = 0;
}

/* Destructor for class sccs_file. */

sccs_file::~sccs_file() {
	if (mode != READ) {
		name.unlock();
	}
	if (mode != CREATE && f != NULL) {
		fclose(f);
	}
}

/* Local variables: */
/* mode: c++ */
/* End: */
