/*
 * sf-delta.c
 *
 * By Ross Ridge
 * Public Domain
 *
 * Members of sccs_file for adding a delta to the SCCS file.
 *
 */

#include "mysc.h"
#include "sccsfile.h"
#include "pfile.h"
#include "seqstate.h"
#include "pipe.h"
#include "run.h"
#include "linebuf.h"

#ifdef CONFIG_NO_STRSTR
#include "strstr.c"
#endif

#ifdef CONFIG_SCCS_IDS
static char const sccs_id[] = "@(#) MySC sf-delta.c 1.10 98/06/06 16:26:00";
#endif

class diff_state_base {
protected:
	int in_lineno, out_lineno;

public:
	enum state { START, NOCHANGE, DELETE, INSERT, END, ERROR, WERROR };

public:
	diff_state_base(): in_lineno(0), out_lineno(0) {}

	virtual enum state process(FILE *out, seq_no seq) = 0;

	virtual char const *get_insert_line() = 0;

	int in_line() { return in_lineno; }
	int out_line() { return out_lineno; }

#ifdef __GNUC__
	virtual ~diff_state_base() {}; // gets rid of a stupid warning
#endif
};

class diff_state: public diff_state_base {
private:
	enum state _state;
	int lines_left;
	int change_left;

	FILE *in;
	class _linebuf linebuf;
	string gname;

	void corrupt();
	void corrupt(char const *msg);

	int next_state();
	int read_line() { return linebuf.read_line(in); }

public:
	diff_state(FILE *f, string name)
		: _state(START), lines_left(0), change_left(0),
	          in(f), gname(name) {}

	enum state process(FILE *out, seq_no seq);

	char const *
	get_insert_line() {
		assert(_state == INSERT);
		assert(linebuf[0] == '>' && linebuf[1] == ' ');
		return linebuf + 2;
	}

};

class null_diff_state: public diff_state_base {
public:
	virtual enum state
	process(FILE *, seq_no) {
		in_lineno++;
		out_lineno++;
		return END;
	}

	virtual char const *get_insert_line() { abort(); }
};

/* Print n appropriate error message when a read operation
   on the diff output fails. */

void
diff_state::corrupt() {
	if (ferror(in)) {
		error(errno, msgptr(59, "%s: Read error."), "(diff pipe)");
	} else {
		error(msgptr(121, "(diff pipe): Unexpected EOF."));
	}
}


/* Print a cryptic error message indicating that something
   is wrong with the diff output. */

void 
diff_state::corrupt(char const *msg) {
	error(msgptr(122, "Diff output corrupt. (%s)"), msg);
}


/* Figure out what the new state should be by processing the
   diff output. */

inline int
diff_state::next_state() {
	if (_state == DELETE && change_left != 0) {
		if (read_line()) {
			corrupt();
			return -1;
		}
		if (strcmp(linebuf, "---\n") != 0) {
			corrupt("---");
			return -1;
		}
		lines_left = change_left;
		change_left = 0;
		_state = INSERT;
		return 0;
	}

	if (_state != NOCHANGE) {
		if (read_line()) {
			if (ferror(in)) {
				corrupt();
				return -1;
			}
			_state = END;
			return 0;
		}

		/* Ignore "\ No newline at end of file" if it appears
		   at the end of the diff output. */

		if (linebuf[0] == '\\') {
			if (!read_line()) {
				corrupt(msgptr(123, "Expected EOF"));
				return -1;
			}
			if (ferror(in)) {
				corrupt();
				return -1;
			}
			_state = END;
			return 0;
		}
			
	}

	char *s = NULL;
	int line1, line2, line3, line4;
	char c;

	line1 = (int) strtol(linebuf, &s, 10);
	line2 = line1;
	if (*s == ',') {
		line2 = (int) strtol(s + 1, &s, 10);
		if (line2 <= line1) {
			corrupt(msgptr(124, "left end line"));
			return -1;
		}
	}

	c = *s;

	assert(c != '\0');
		
	if (c == 'a') {
		if (line1 >= in_lineno) {
			_state = NOCHANGE;
			lines_left = line1 - in_lineno + 1;
			return 0;
		}
		if (line1 + 1 != in_lineno) {
			corrupt(msgptr(125, "left start line"));
			return -1;
		}
	} else {
		if (line1 > in_lineno) {
			_state = NOCHANGE;
			lines_left = line1 - in_lineno;
			return 0;
		}
		if (line1 != in_lineno) {
			corrupt(msgptr(125, "left start line"));
			return -1;
		}
	}

	line3 = (int) strtol(s + 1, &s, 10);
	if (c == 'd') {
		if (line3 != out_lineno) {
			corrupt(msgptr(126, "right start line"));
			return -1;
		}
	} else {
		if (line3 != out_lineno + 1) {
			corrupt(msgptr(126, "right start line"));
			return -1;
		}
	}

	line4 = line3;
	if (*s == ',') {
		line4 = (int) strtol(s + 1, &s, 10);
		if (line4 <= line3) {
			corrupt(msgptr(127, "right end line"));
			return -1;
		}
	}

	if (*s != '\n') {
		corrupt(msgptr(128, "EOL"));
		return -1;
	}

	switch(c) {
	case 'a':
		_state = INSERT;
		lines_left = line4 - line3 + 1;
		break;		

	case 'd':
		_state = DELETE;
		lines_left = line2 - line1 + 1;
		break;

	case 'c':
		_state = DELETE;
		lines_left = line2 - line1 + 1;
		change_left = line4 - line3 + 1;
		break;

	default:
		corrupt(msgptr(129, "unknown operation"));
		return -1;
	}

	return 0;
}


/* Figure out whether a line is being inserted, deleted or left unchanged.
   Output new control lines accordingly. */

inline /* enum */ diff_state::state
diff_state::process(FILE *out, seq_no seq) {
	if (_state != INSERT) {
		in_lineno++;
	}

	if (_state != END) {
		assert(lines_left >= 0);
		if (lines_left == 0) {
			if (_state == DELETE || _state == INSERT) {
				if (fprintf(out, "\001E %d\n", seq) == EOF) {
					return WERROR;
				}
			}
			if (next_state() == -1) {
				return ERROR;
			}

			if (_state == INSERT) {
				if (fprintf(out, "\001I %d\n", seq) == EOF) {
					return WERROR;
				}
			} else if (_state == DELETE) {
				if (fprintf(out, "\001D %d\n", seq) == EOF) {
					return WERROR;
				}
			}
		}
		lines_left--;
	}

	if (_state == DELETE) {
		if (read_line()) {
			corrupt();
			return ERROR;
		}
		if (linebuf[0] != '<' || linebuf[1] != ' ') {
			corrupt("<");
			return ERROR;
		}
	} else {
		if (_state == INSERT) {
			if (read_line()) {
				corrupt();
				return ERROR;
			}
			if (linebuf[0] != '>' || linebuf[1] != ' ') {
				corrupt(">");
				return ERROR;
			}
		}
		out_lineno++;
	}

	return _state;
}


/* Warns or quits if the new delta doesn't include any id keywords */

int
sccs_file::check_keywords_in_file(char const *name) {
	FILE *f = fopen(name, "r");
	if (f == NULL) {
		error(errno, msgptr(53, "%s: Can't open file for reading."),
		      name);
		return -1;
	}

	if (flags.id_keywords && *(char const *)flags.id_keywords != '\0') {
		while(!read_line_param(f)) {
			if (strstr(linebuf, flags.id_keywords) != NULL) {
				fclose(f);
				return 0;
			}
		}
		error(msgptr(118, "%s: Required keywords \"%s\" missing."),
		      name, (char const *) flags.id_keywords);
		return -1;
	}

	while(!read_line_param(f)) {
		if (check_id_keywords(linebuf)) {
			fclose(f);
			return 0;
		}
	}

	if (flags.id_keywords) {
		error(msgptr(130, "%s: No id keywords."), name);
		return -1;
	}

	fprintf(stderr, msgptr(119, "%s: Warning: No id keywords.\n"), name);
	fclose(f);

	return 0;
}


class diff_state_base *
sccs_file::create_diff_state(string gname, seq_state &sstate,
			     class pipe &diff_out) {
	if (gname == NULL) {
		return new null_diff_state();
	}

	class pipe diff_in;
	
	FILE *get_out = diff_in.write_stream();
	if (get_out != NULL) {
		struct subst_parms parms(get_out, NULL, delta(), 
					 0, sccs_date(NULL));
#ifdef CONFIG_NO_PIPE
		seq_state gsstate = sstate;
		if (get(msgptr(131, "(diff pipe)"), gsstate, parms) == -1) {
			diff_in.set_failed();
		}
#else
		sccs_file file(name, READ);
		if (file.failed()
		    || file.get(msgptr(131, "(diff pipe)"),
				sstate, parms) == -1) {
			diff_in.set_failed();
		}
#endif
	}
	
	diff_in.write_close();
	
	int ret;
	list<string> args;

#ifdef CONFIG_DIFF_SWITCHES
	args.add(CONFIG_DIFF_SWITCHES);
#endif

#ifdef CONFIG_NO_PIPE

	ret = run_diff(gname, &diff_in, &diff_out, args);
	if (ret != STATUS(0) && ret != STATUS(1)) {
		error(msgptr(132, "%s: Command failed.  (status = %d)"),
		      CONFIG_DIFF_COMMAND, STATUS_MSG(ret));
		diff_in.set_failed();
		diff_out.set_failed();
	}

	diff_in.read_close();

#else /* CONFIG_NO_PIPE */

	run_diff(gname, &diff_in, &diff_out, args);
	
	ret = diff_in.read_close();
	if (ret != STATUS(0)) {
		error(msgptr(133, "get: Subprocess exited abnormally."
			     "  (status = %d)"), STATUS_MSG(ret));
		diff_in.set_failed();
		diff_out.set_failed();
	}

#endif /* CONFIG_NO_PIPE */

	diff_out.write_close();

	if (diff_in.get_failed() || diff_out.get_failed()) {
		return NULL;
	}

	return new diff_state(diff_out.read_stream(), gname);
}

/* Adds a new delta to the SCCS file.  It doesn't add the delta to the
   delta list in sccs_file object, so this should be the last operation
   performed before the object is destroyed. */

int
sccs_file::add_delta(string gname, sccs_pfile &pfile,
		     list<string> mrs, list<string> comments) {
	assert(mode == UPDATE);

	if (gname != NULL) {
		if (check_keywords_in_file(gname) == -1) {
			return -1;
		}
	}

	seq_state sstate(delta_table.highest_seqno());
	struct delta const *got_delta = delta_table.find(pfile->got);
	if (got_delta == NULL) {
		error(msgptr(134, "Locked delta doesn't exist!"));
		return -1;
	}

	if (prepare_seqstate(sstate, got_delta->seq) == -1) {
		return -1;
	}

	prepare_seqstate(sstate, pfile->include, pfile->exclude, 
			 sccs_date(NULL));

	class pipe diff_out;
	class diff_state_base *dstate = create_diff_state(gname, sstate,
							  diff_out);

	if (dstate == NULL || seek_to_body() == -1) {
		return -1;
	}


	list<seq_no> included, excluded;
	seq_no seq;
	for(seq	= 1; seq < delta_table.highest_seqno(); seq++) {
		if (sstate.is_explicit(seq)) {
			if (sstate.is_included(seq)) {
				included.add(seq);
			} else if (sstate.is_excluded(seq)) {
				excluded.add(seq);
			}
		}
	}

	struct delta new_delta('D', pfile->delta,
			       sccs_date::now(), get_user_name(),
			       delta_table.highest_seqno() + 1, got_delta->seq,
			       included, excluded, mrs, comments);
	new_delta.inserted = 0;
	new_delta.deleted = 0;
	new_delta.unchanged = 0;


	FILE *out = start_update(new_delta);

#undef DEBUG_FILE
#ifdef DEBUG_FILE
	FILE *df = fopen("delta.dbg", "w");
#endif

	while(1) {
		int c = read_line();
		if (c == -2) {
			fclose(out);
			delete dstate;
			return -1;
		}

		if (c != 0 && c != -1) {
			int err = 0;
			seq_no seq = strict_atous(linebuf + 3, &err);
			if (err || seq < 1 
			    || seq > delta_table.highest_seqno()) {
				corrupt(msgptr(135, "Invalid sequence number"));
				fclose(out);
				delete dstate;
				return -1;
			}

			char const *msg = NULL;

			switch(c) {
			case 'E':
				msg = sstate.end(seq);
				break;

			case 'D':
			case 'I':
				msg = sstate.start(seq, c == 'I');
				break;

			default:
				msg = msgptr(136, "Unexpected control line");
				break;
			}

			if (msg != NULL) {
				corrupt(msg);
				fclose(out);
				delete dstate;
				return -1;
			}
		} else if (sstate.include_line()) {
			/* enum */ diff_state::state action;

			do {
				action = dstate->process(out, new_delta.seq);

				switch(action) {
				case diff_state::DELETE:
					assert(c != -1);
#ifdef DEBUG_FILE
					fprintf(df, "%4d %4d - %s\n",
						dstate->in_line(),
						dstate->out_line(),
					        (char const *) linebuf);
						fflush(df);
#endif
					new_delta.deleted++;
					break;

				case diff_state::INSERT:
#ifdef DEBUG_FILE
					fprintf(df, "%4d %4d + %s",
						dstate->in_line(),
						dstate->out_line(),
					        dstate->get_insert_line());
						fflush(df);
#endif
					new_delta.inserted++;
					if (fputs(dstate->get_insert_line(),
						  out) == EOF) {
						xfile_error(msgptr(111, "Write error."));
						delete dstate;
						fclose(out);
						return -1;
					}
					break;

				case diff_state::END:
					if (c == -1) {
						break;
					}
					/* FALLTHROUGH */
				case diff_state::NOCHANGE:
					assert(c != -1);
#ifdef DEBUG_FILE
					fprintf(df, "%4d %4d   %s\n",
						dstate->in_line(),
						dstate->out_line(),
					        (char const *) linebuf);
					fflush(df);
#endif
					new_delta.unchanged++;
					break;

				case diff_state::WERROR:
					xfile_error(msgptr(111,
							   "Write error."));
					/* FALLTHROUGH */
				case diff_state::ERROR:
					delete dstate;
					fclose(out);
					return -1;

				default:
					abort();
				}

			} while(action == diff_state::INSERT);
		}

		if (c == -1) {
			break;
		}

		if (fputs(linebuf, out) == EOF || putc('\n', out) == EOF) {
			xfile_error(msgptr(111, "Write error."));
			delete dstate;
			fclose(out);
			return -1;
		}
	}

	delete dstate;

#ifdef DEBUG_FILE
	fclose(df);
#endif

	if (gname != NULL) {
		int ret;

#ifdef CONFIG_NO_PIPE
		diff_out.read_close();
#else
		ret = diff_out.read_close();
		if (ret != STATUS(0) && ret != STATUS(1)) {
			error(msgptr(132, "%s: Command failed."
				     "  (status = %d)"),
			      CONFIG_DIFF_COMMAND, STATUS_MSG(ret));
			fclose(out);
			return -1;
		}
#endif
	}

	pfile.delete_lock();
	if (pfile.update() == -1) {
		fclose(out);
		return -1;
	}

	return end_update(out, new_delta);
}

/* Local variables: */
/* mode: c++ */
/* End: */

