/*
 * sf-get.c
 *
 * By Ross Ridge
 * Public Domain
 *
 * Members of class sccs_file used in getting deltas. 
 * 
 */

#include "mysc.h"
#include "sccsfile.h"
#include "pfile.h"
#include "seqstate.h"

#ifdef CONFIG_NO_STRSTR
#include "strstr.c"
#endif

#ifdef CONFIG_SCCS_IDS
static char const sccs_id[] = "@(#) MySC sf-get.c 1.7 98/07/01 19:51:05";
#endif

int
sccs_file::prepare_seqstate(seq_state &state, seq_no seq) {
	while(seq != 0) {
		if (state.is_predecessor(seq)) {
			error(msgptr(137, "%s: Loop in deltas."),
			      (char const *) name);
			return -1;
		}
		state.set_predecessor(seq);

		int len;
		int i;

		struct delta const &delta = delta_table[seq];

		len = delta.included.length();
		for(i = 0; i < len; i++) {
			state.pred_include(delta.included[i]);
		} 		

		len = delta.excluded.length();
		for(i = 0; i < len; i++) {
			state.pred_exclude(delta.excluded[i]);
		} 		

		seq = delta.prev_seq;
	}
	return 0;
}

int
sccs_file::get(string gname, class seq_state &state, struct subst_parms &parms,
	       int (sccs_file::*subst_fn)(char const *,
					  struct subst_parms *) const,
	       int show_sid, int show_module) {
	int c;

	assert(mode != CREATE);
	
	char const *req_idkey = flags.id_keywords;
	if (req_idkey != NULL && req_idkey[0] == '\0') {
		req_idkey = NULL;
	}

	if (seek_to_body() == -1) {
		return -1;
	}

	/* "@[DI] n" should start the body of the SCCS file */

	c = read_line();
	if (c != 'I' && c != 'D') {
		corrupt(msgptr(139, "First delta missing?!?"));
		return -1;
	}

	if (check_arg() == -1) {
		return -1;
	}

	int err = 0;
	seq_no seq = strict_atous(linebuf + 3, &err);
	state.start(seq, c == 'I');
	if (err || seq < 1 || seq > delta_table.highest_seqno()) {
		corrupt(msgptr(140, "Invalid serial number"));
	}

	FILE *out = parms.out;

	while(1) {
		c = read_line();

		if (c == -2) {
			return -1;
		}

		if (c == -1) { 
			/* EOF */
			break;
		}

		if (c == 0) {
			/* A non-control line */

			if (!state.include_line()) {
				continue;
			}

			parms.out_lineno++;

			if (show_module) {
				string module = get_module_name();
				if (fprintf(out, "%s\t",
					    (char const *) module) == EOF) {
					error(errno, 
					      msgptr(42, "%s: Write error."),
					      (char const *) gname);
					return -1;
				}
			}

			if (show_sid) {
				if (delta_table[state.active_seq()]
				    .id.print(out) == EOF
				    || putc('\t', out) == EOF) {
					error(errno,
					      msgptr(42, "%s: Write error."), 
					      (char const *) gname);
					return -1;
				}
			}

			int err;
			if (subst_fn != NULL) {
				err = (this->*subst_fn)(linebuf, &parms);
			} else {
				err = fputs(linebuf, out);
				if (req_idkey == NULL && !parms.found_id 
				    && check_id_keywords(linebuf)) {
					parms.found_id = 1;
				}
			} 

			if (!parms.found_id && req_idkey != NULL
			    && strstr(linebuf, req_idkey) != NULL) {
				parms.found_id = 1;
				req_idkey = NULL;
			}
				
			if (err == EOF || fputc('\n', out) == EOF) {
				error(errno, msgptr(42, "%s: Write error."),
				      (char const *) gname);
				return -1;
			}
			continue;
		}

		/* A control line */

		if (check_arg() == -1) {
			return -1;
		}

		int err = 0;
		seq_no seq = strict_atous(linebuf + 3, &err);
		if (err || seq < 1 || seq > delta_table.highest_seqno()) {
			corrupt(msgptr(140, "Invalid serial number"));
			return -1;
		}

		char const *msg = NULL;

		switch(c) {
		case 'E':
			msg = state.end(seq);
			break;

		case 'D':
		case 'I':
			msg = state.start(seq, c == 'I');
			break;

		default:
			msg = msgptr(136, "Unexpected control line");
		}

		if (msg != NULL) {
			corrupt(msg);
			return -1;
		}
	}

	if (fflush(out) == EOF) {
		error(errno, msgptr(42, "%s: Write error."),
		      (char const *) gname);
		return -1;
	}

	return 0;
}

/* Local variables: */
/* mode: c++ */
/* End: */
