/*
 * sf-prs.c
 *
 * By Ross Ridge
 * Public Domain
 *
 * Members of the class sccs_file for printing selected parts of an
 * SCCS file.
 *
 */

#include "mysc.h"
#include "sccsfile.h"
#include "seqstate.h"

#ifdef CONFIG_SCCS_IDS
static char const sccs_id[] = "@(#) MySC sf-prs.c 1.8 98/04/21 16:44:20";
#endif

inline int
sccs_file::print_get(FILE *out, string name, seq_no seq) {
	struct subst_parms parms(out, NULL, delta(), 0, sccs_date(NULL));

	class seq_state state(delta_table.highest_seqno());

	if (prepare_seqstate(state, seq) == -1) {
		return EOF;
	}

	if (get(name, state, parms) == -1) {
		errno = ENOEXEC;
		return EOF;
	}
	return 0;
}

/* Prints a list of sequence numbers on the same line. */

static int
print_seq_list(FILE *out, list<seq_no> const &list) {
	int i;
	int len = list.length();

	if (len > 0) {
		if (fprintf(out, "%u", list[0]) == EOF) {
			return EOF;
		}
		for(i = 1; i < len; i++) {
			if (fprintf(out, " %u", list[i]) == EOF) {
				return EOF;
			}
		}
	}
	return 0;
}


/* Prints a list of strings, one per line. */

static int
print_string_list(FILE *out, list<string> const &list) {
	int i;
	int len = list.length();

	for(i = 0; i < len; i++) {
		if (fprintf(out, "%s\n", (char const *) list[i]) == EOF) {
			return EOF;
		}
	}
	return 0;
}


/* Prints a string flag with it's name. */

static int
print_flag2(FILE *out, char const *s, string it) {
	if (it != NULL) {
		return fprintf(out, "%s: %s\n", s, (char const *) it);
	}
	return 0;
}


/* Prints a boolean flag with it's name. */

static int
print_flag2(FILE *out, char const *s, int it) {
	if (it) {
		return fprintf(out, msgptr(145, "%s: yes\n"), s);
	}
	return 0;
}


/* Prints a flag whose type has a print(FILE *) member with it's name. */

template <class TYPE>
int
print_flag2(FILE *out, char const *s, TYPE it) {
	if (it != NULL) {
		if (fprintf(out, "%s: ", s) == EOF
		    || it.print(out) == EOF
		    || putc('\n', out) == EOF) {
			return EOF;
		}
	}
	return 0;
}

#if 0
template <class TYPE>
inline int
print_flag2(FILE *out, char *s, TYPE it) {
	return print_flag2(out, (char const *) s, it);
}
#endif

/* Prints all the flags of an SCCS file. */

int
sccs_file::print_flags(FILE *out) const {
	if (print_flag2(out, msgptr(146, "Module Type"), flags.type) == EOF
	    || print_flag2(out, msgptr(147, "MR Validation"),
			   flags.mr_checker) == EOF
	    || print_flag2(out, msgptr(148, "Keyword Validation"),
			   flags.id_keywords) == EOF
	    || print_flag2(out, msgptr(149, "Branch"), flags.branch) == EOF
	    || print_flag2(out, msgptr(150, "Module Name"),
			   flags.module) == EOF
	    || print_flag2(out, msgptr(151, "Floor"), flags.floor) == EOF
	    || print_flag2(out, msgptr(152, "Ceiling"), flags.ceiling) == EOF
	    || print_flag2(out, msgptr(153, "Default SID"),
			   flags.default_sid) == EOF
	    || print_flag2(out, msgptr(154, "Null Deltas"),
			   flags.null_deltas) == EOF
	    || print_flag2(out, msgptr(155, "Joint Editing"),
			   flags.joint_edit) == EOF) {
		return EOF;
	}
	int r;
	if (flags.all_locked) {
		r = fputs(msgptr(156, "Locked Releases: a"), out);
	} else {
		r = print_flag2(out, msgptr(157, "Locked Releases"),
				flags.locked);
	}
	if (r == EOF) {
		return EOF;
	}
	return print_flag2(out, msgptr(158, "User Keyword"), flags.user_def);
}


/* Prints "yes" or "no" according to the value of a boolean flag. */

static int
print_yesno(FILE *out, int flag) {
	if (flag) {
		return fputs(msgptr(159, "yes"), out);
	} else {
		return fputs(msgptr(160, "no"), out);
	}
}


/* Prints the the value of string flag. */

template <class TYPE>
int
print_flag(FILE *out, TYPE it) {
	if (it == NULL) {
		return fputs("none", out);
	} else {
		return it.print(out);
	}
}


/* Prints the value of flag whose type has a print(FILE *) member. */

static int
print_flag(FILE *out, string s) {
	if (s == NULL) {
		return fputs("none", out);
	} else {
		return fputs(s, out);
	}
}


/* These macros are used to convert the one or two characters a prs
   data keyword in an unsigned value used in the switch statement 
   in the function below. */

#define KEY1(c)		((unsigned char)(c))
#define KEY2(c1, c2)	(((unsigned char)(c1)) * 256 + (unsigned char)(c2))

/* Prints selected parts of an SCCS file and the specified entry in the
   delta table. */

int
sccs_file::print_delta(FILE *out, char const *format,
		       struct delta const &delta) {
	char const *s = format;
	int error_status = 0;

	while(error_status != EOF) {
		char c = *s++;

		if (c == '\0') {
			break;
		}

		if (c != ':' || s[0] == '\0') {
			error_status = putc(c, out);
			continue;
		}

		char const *back_to = s;
		unsigned key = 0;

		if (s[1] == ':') {
			key = KEY1(s[0]);
			s += 2;
		} else if (s[2] == ':') {
			key = KEY2(s[0], s[1]);
			s += 3;
		} else {
			error_status = putc(':', out);
			continue;
		}

		switch(key) {
		default:
			s = back_to;
			error_status = putc(':', out);
			continue;
			
		case KEY2('D','t'):
			error_status 
				= print_delta(out,
					      ":DT: :I: :D: :T: :P: :DS: :DP:",
					      delta);
			break;

		case KEY2('D','L'):
			error_status = print_delta(out, ":Li:/:Ld:/:Lu:",
						   delta);
			break;

		case KEY2('L','i'):
			error_status = fprintf(out, "%05u", delta.inserted);
			break;

		case KEY2('L','d'):
			error_status = fprintf(out, "%05u", delta.deleted);
			break;

		case KEY2('L','u'):
			error_status = fprintf(out, "%05u", delta.unchanged);
			break;

		case KEY2('D','T'):
			error_status = putc(delta.type, out);
			break;

		case KEY1('I'):
			error_status = delta.id.print(out);
			break;

		case KEY1('R'):
			error_status = delta.id.printf(out, 'R');
			break;

		case KEY1('L'):
			error_status = delta.id.printf(out, 'L');
			break;

		case KEY1('B'):
			error_status = delta.id.printf(out, 'B');
			break;

		case KEY1('S'):
			error_status = delta.id.printf(out, 'S');
			break;

		case KEY1('D'):
			error_status = delta.date.printf(out, 'D');
			break;

		case KEY2('D','y'):
			error_status = delta.date.printf(out, 'y');
			break;

		case KEY2('D','m'):
			error_status = delta.date.printf(out, 'o');
			break;

		case KEY2('D','d'):
			error_status = delta.date.printf(out, 'd');
			break;

		case KEY1('T'):
			error_status = delta.date.printf(out, 'T');
			break;

		case KEY2('T','h'):
			error_status = delta.date.printf(out, 'h');
			break;

		case KEY2('T','m'):
			error_status = delta.date.printf(out, 'm');
			break;

		case KEY2('T','s'):
			error_status = delta.date.printf(out, 's');
			break;

		case KEY1('P'):
			error_status = fputs(delta.user, out);
			break;

		case KEY2('D','S'):
			error_status = fprintf(out, "%u", delta.seq);
			break;

		case KEY2('D','P'):
			error_status = fprintf(out, "%u", delta.prev_seq);
			break;

		case KEY2('D', 'I'):
			error_status = print_delta(out, ":Dn:/:Dx:/:Dg:",
						   delta);
			break;

		case KEY2('D','n'):
			error_status = print_seq_list(out, delta.included);
			break;

		case KEY2('D','x'):
			error_status = print_seq_list(out, delta.excluded);
			break;

		case KEY2('D','g'):
			error_status = print_seq_list(out, delta.ignored);
			break;

		case KEY2('M','R'):
			error_status = print_string_list(out, delta.mrs);
			break;

		case KEY1('C'):
			error_status = print_string_list(out, delta.comments);
			break;

		case KEY2('U','N'):
			error_status = print_string_list(out, users);
			break;

		case KEY2('F', 'L'):
			error_status = print_flags(out);
			break;
			
		case KEY1('Y'):
			error_status = print_flag(out, flags.type);
			break;
			
		case KEY2('M','F'):
			error_status = print_yesno(out,
						   flags.mr_checker != NULL);
			break;

		case KEY2('M','P'):
			error_status = print_flag(out, flags.mr_checker);
			break;
			
		case KEY2('K','F'):
			error_status = print_yesno(out,
						   flags.id_keywords != NULL);
			break;

		case KEY2('K','V'):
			error_status = print_flag(out, flags.id_keywords);
			break;

		case KEY2('B','F'):
			error_status = print_yesno(out, flags.branch);
			break;

		case KEY1('J'):
			error_status = print_yesno(out, flags.joint_edit);
			break;
			
		case KEY2('L','K'):
			if (flags.all_locked) {
				error_status = putc('a', out);
			} else {
				error_status = print_flag(out, flags.locked);
			}
			break;

		case KEY1('Q'):
			error_status = print_flag(out, flags.user_def);
			break;

		case KEY1('M'):
			error_status = print_flag(out, get_module_name());
			break;
			
		case KEY2('F','B'):
			error_status = print_flag(out, flags.floor);
			break;
			
		case KEY2('C','B'):
			error_status = print_flag(out, flags.ceiling);
			break;

		case KEY2('D','s'):
			error_status = print_flag(out, flags.default_sid);
			break;

		case KEY2('N','D'):
			error_status = print_yesno(out, flags.null_deltas);
			break;

		case KEY2('F','D'):
			error_status = print_string_list(out, comments);
			break;

		case KEY2('B','D'):
			if (seek_to_body() == -1) {
				error_status = EOF;
				break;
			}
			while(1) {
				int c = read_line();
				if (c == -1) {
					break;
				}
				if (c == -2
				    || fputs(linebuf, out) == EOF
				    || putc('\n', out) == EOF) {
					error_status = EOF;
					break;
				}
			}
			break;

		case KEY2('G','B'):
			error_status = print_get(out, "-", delta.seq);
			break;

		case KEY1('W'):
			error_status = print_delta(out, ":Z::M:\t:I:", delta);
			break;

		case KEY1('A'):
			error_status = print_delta(out, ":Z::Y: :M: :I::Z:",
						   delta);
			break;

		case KEY1('Z'):
			if (fputc('@', out) == EOF
			    || fputs("(#)", out) == EOF) {
				error_status = EOF;
			}
			break;

		case KEY1('F'):
			error_status = fputs(base_part(name), out);
			break;

		case KEY2('P','N'):
			error_status = fputs(name, out);
			break;
		}
	}
	return error_status;
}

/* Find the SID of the most recently created delta with the same release
   and level as the requested SID. */
   
sid
sccs_file::find_most_recent_sid(sid id) const {
	delta_iterator iter(delta_table);

#if 0
	fputs("find_most_recented_sid(", stderr);
	id.dprint(stderr);
	fputs(")\n", stderr);
#endif

	while(iter.next()) {
		if (id.partial_match(iter->id)) {

#if 0
			fputs("match: ", stderr);
			iter->id.dprint(stderr);
			putc('\n', stderr);
#endif
			return iter->id;
		}
	}
	return NULL;
}

/* Prints out parts of the SCCS file.  */

int
sccs_file::prs(FILE *out, string format, sid rid, sccs_date cutoff,
	       enum when when, int all_deltas) {
	if (rid.partial_sid()) {
		rid = find_most_recent_sid(rid);
	}

	if (when != SIDONLY && !cutoff.valid()) {
		struct delta const *delta = delta_table.find(rid);
		if (delta == NULL) {
			error(msgptr(120, "%s: Requested SID doesn't exist."),
			      (char const *) name);
			return -1;
		}
		cutoff = delta->date;
	}

	delta_iterator iter(delta_table);
	while(iter.next(all_deltas)) {
		switch(when) {
		case EARLIER:
			if (iter->date > cutoff) {
				continue;
			}
			break;

		case SIDONLY:
			if (rid == iter->id) {
				break;
			}
			continue;

		case LATER:
			if (iter->date < cutoff) {
				continue;
			}
			break;
		}

		if (print_delta(out, format, *iter.operator->()) == EOF
		    || putc('\n', out) == EOF) {
			if (errno != ENOEXEC) {
				error(errno, msgptr(42, "%s: Write error."),
				      "-");
			}
			return -1;
		}
	}
	return 0;
}

/* Local variables: */
/* mode: c++ */
/* End: */
